﻿/*
 * Created by SharpDevelop.
 * User: phil
 * Date: 2010.08.17.
 * Time: 15:20
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
//
using TetheredSun.Mathematics;


namespace TetheredSun.Stochastic
{
	/// <summary>
	/// Description of PowerLawRandomGenerator.
	/// </summary>
	public class PowerLawRandomGenerator : IRandomGenerator
	{
		protected double exponent;
		protected double minimum;
		protected double maximum;
		protected double a;
		protected double b;
		protected double c;
		protected Random uniform;
		protected bool exponentOne;
		
		public PowerLawRandomGenerator(double exponent, double minimum, double maximum, int seed)
		{
			if (exponent <= 0) {
				Exception exception = new ArgumentException("The exponent cannot be zero or negative.");
				exception.Data.Add("Exponent", exponent);
				throw exception;
			}
			
			if (maximum <= minimum) {
				Exception exception = new ArgumentException("The maximum cannot be less than or equal to the minimum.");
				exception.Data.Add("Minimum", minimum);
				exception.Data.Add("Maximum", maximum);
				throw exception;
			}
			this.exponent = exponent;
			this.minimum = minimum;
			this.maximum = maximum;
			uniform = (seed < 0) ? new Random() : new Random(seed);
			// Derived constants:
			exponentOne = (this.exponent == 1.0);
			if (!exponentOne) {
				a = Math.Pow(minimum, 1.0 - exponent);
				b = Math.Pow(maximum, 1.0 - exponent);
				c = (1.0 - exponent) / (b - a);
			} else {
				a = Math.Log(minimum);
				b = Math.Log(maximum);
				c = 1.0 / (b - a);
			}
		}
		
		public PowerLawRandomGenerator(double exponent, double minimum, double maximum) :
			this(exponent, minimum, maximum, -1) { }
		
		public double ExpectedValue {
			get {
				if (exponentOne) {
					return c * (maximum - minimum);
				} else if (exponent == 2.0) {
					return c * (Math.Log(maximum) - Math.Log(minimum));
				} else {
					return c / (2.0 - exponent) * (b * maximum - a * minimum);
				}
			}
		}
		
		
		public double Next()
		{
			double u = uniform.NextDouble();
			
			return exponentOne ?
				Math.Pow(maximum, u) / Math.Pow(minimum, u - 1.0) :
				Math.Pow(u * (b - a) + a, 1.0 / (1.0 - exponent));
		}
		
		public double[] GenerateSequence(int count)
		{
			double[] sequence = new double[count];
			
			for (int i = 0; i < count; i++) {
				sequence[i] = Next();
			}
			
			return sequence;
		}
		
		public double ProbabilityDensity(double x)
		{
			return (minimum < x) && (maximum > x) ?
				c / Math.Pow(x, exponent) :
				0.0;
		}
		
		public static double GetExpectedValue(double k, double min, double max)
		{
			double A;
			double B;
			double C;
			
			if (k == 1.0) {
				A = Math.Log(min);
				B = Math.Log(max);
				C = 1.0 / (B - A);
				return C * (max - min);
			} else {
				A = Math.Pow(min, 1.0 - k);
				B = Math.Pow(max, 1.0 - k);
				C = (1.0 - k) / (B - A);
				if (k == 2.0) {
					return C * (Math.Log(max) - Math.Log(min));
				} else {
					return C / (2.0 - k) * (B * max - A * min);
				}
			}
		}
		
		public static double GetMinimum(double exponent, double maximum, double expectedValue)
		{
			double[] parameters = new double[4];
			
			parameters[0] = exponent;
			parameters[2] = maximum;
			parameters[3] = expectedValue;
			
			return NumericEquationSolver.SolveBySecant(GetExpectedValueDifference,
			                                           parameters,
			                                           1,
			                                           0.001,
			                                           maximum + 1.0);
		}
		
		public static double GetMaximum(double exponent, double minimum, double expectedValue)
		{
			double[] parameters = new double[4];
			
			parameters[0] = exponent;
			parameters[1] = minimum;
			parameters[3] = expectedValue;
			
			return NumericEquationSolver.SolveBySecant(GetExpectedValueDifference,
			                                           parameters,
			                                           2,
			                                           minimum + 0.001,
			                                           minimum + 4.0 * expectedValue);
		}
		
		private static double GetExpectedValueDifference(double[] parameters)
		{
			double k = parameters[0];
			double min = parameters[1];
			double max = parameters[2];
			double expectedValue = parameters[3];
			
			return GetExpectedValue(k, min, max) - expectedValue;
		}
	}
}
